# Copyright Justin R. Goheen.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC

import numpy as np

import ib_insync as ib


class ConnectionMixins(ABC):
    """an extension of ib_insync.connection.Connection"""

    def __init__(self):
        super().__init__()
        self._clientid = np.random.randint(0, 100000)

    @property
    def ports(self):
        _ports = dict(
            tws=dict(paper=7497, live=7496),
            gateway=dict(paper=4002, live=4001),
        )
        return _ports

    @property
    def port(self):
        return self.ports[self.platform][self.connection_type]

    @property
    def host(self):
        return "127.0.0.1"

    @property
    def clientid(self):
        return self._clientid

    def connect(self):
        self.app.connect(
            self.host,
            self.port,
            clientId=self.clientid,
        )

    def disconnect(self):
        self.app.disconnect()

    def controller(self, userid, password, ib_major_version: str, ibc_path: str, ibc_ini: str):

        ibc = ib.ibcontroller.IBC(
            ib_major_version,
            gateway=False,
            tradingMode="live",
            userid=userid,
            password=password,
            ibcPath=ibc_path,
            ibcIni=ibc_ini,
        )

        ibc.start()
        self._ib.run()



# Copyright Justin R. Goheen.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from abc import ABC
from dataclasses import dataclass

from bs4 import BeautifulSoup


@dataclass
class AdvisorDataTypes:
    groups: int = 1
    profiles: int = 2
    account_aliases: int = 3


@dataclass
class AdvisorGroup:
    members: tuple = ()


class AdvisorMixins(ABC):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _get_group_members(self, group_soup):
        group = AdvisorGroup()
        for member in group_soup.find_all("string"):
            group.members += (member.text,)
        return group

    def _groups(self):
        group = self.app.requestFA(AdvisorDataTypes.groups)
        soup = BeautifulSoup(group, features="lxml")
        groups = soup.find_all("group")
        return groups

    @property
    def groups(self):
        groups = {i.find("name").text: self._get_group_members(i) for i in self._groups}
        return groups

    def get_group(self, group_name):
        return self.groups[group_name]



# Copyright Justin R. Goheen.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import datetime
import os
import sys
from typing import Dict, Optional

import pytz
from rich import print as rprint

import ib_insync as ib
from ibquant.core import AppBase


class Trader(AppBase):
    """
    a base class for Trading Agents
    """

    def __init__(
        self,
        platform: str,
        connection_type: str,
        contract_type: str,
        account: Optional[str] = None,
        symbol: Optional[str] = None,
        exchange: Optional[str] = None,
        expiry: Optional[str] = None,
        trailing_limit: bool = False,
        trail_type: Optional[str] = None,
        price_source: str = "close",
        order_size: int = 1,
        stop_offset: Optional[int] = None,
        time_sentinel: Optional[int] = None,
        backfill_history: bool = False,
        persist_history: bool = False,
        data_dir: Optional[str] = None,
        logs_dir: Optional[str] = None,
        tws_timezone: Optional[str] = None,
    ) -> None:
        super().__init__(
            platform=platform, connection_type=connection_type, account=account, contract_type=contract_type
        )

        self.validate_contract_config()

        self.contract = self.contract_method(
            symbol=symbol or "", exchange=exchange or "", lastTradeDateOrContractMonth=expiry or ""
        )

        self.trailing_limit = trailing_limit
        self.trail_type = trail_type
        self.order_size = order_size
        self.time_sentinel = time_sentinel
        self._stop_offset = stop_offset
        self.price_source = price_source
        self.data_dir = data_dir
        self.logs_dir = logs_dir
        self.app.TimezoneTWS = pytz.timezone(tws_timezone) if tws_timezone is not None else None

        if backfill_history:
            rprint("\n" + f"[bold green][FETCHING HISTORY][/bold green] {datetime.datetime.now()}")
            # TODO add rich progress bar
            self.history = self.historical_bars(
                end_date_time="",
                duration="1 D",
                bar_size="5 secs",
                what_to_show="TRADES",
                use_rth=True,
            )

            self.history = self.to_dataframe(self.history)

            if persist_history:
                self.persist_history_to_logs(
                    data_dir=f"{os.path.join('data', f'history_{datetime.datetime.now().date().csv}')}"
                )

    @property
    def offset(self) -> Dict[str, str]:
        return dict(stop=self._stop_offset)

    @property
    def open_orders(self) -> bool:
        trades = [getattr(self, attr) for attr in ["parent_trade", "trailing_oder"] if hasattr(self, attr)]
        return any(i.isActive() for i in trades)

    @property
    def no_position(self) -> bool:
        """determines if position is 0"""
        return self.position == 0

    @property
    def pending_parentorder(self) -> bool:
        """determines if a parent order associated"""
        if hasattr(self, "parent_trade"):
            return self.parent_trade.isActive() and self.no_position
        else:
            return False

    @property
    def start_trading_ops(self) -> bool:
        """sets a logic to trade if position is 0 when first data message arrives"""

    @property
    def flank(self):
        """gets and sets the stops and take-profits"""

    @property
    def any_flank_cancelled(self) -> bool:
        """used by flanked_cancelled to check if flanks are active"""
        trades = [getattr(self, attr) for attr in self.flanks if hasattr(self, attr)]
        return any(not trade.isActive() for trade in trades)

    @property
    def flank_cancelled(self) -> bool:
        """sets a logic to determin if stops or take-profits have been cancelled while parent is active"""
        if hasattr(self, "parent_trade"):
            parent_trade_active_or_filled = (
                self.parent_trade.isActive() or self.parent_trade.orderStatus.status == "Filled"
            )
            return parent_trade_active_or_filled and not self.no_position and self.any_flank_cancelled
        return False

    @property
    def flank_filled(self) -> bool:
        """determines if a stop or take-profit has been filled"""

    @property
    def reentry(self) -> bool:
        """determines if Trader should place a new trade after previous trade is filled and position is 0"""

    @property
    def continue_trading_ops(self) -> bool:
        """determines if trades have been placed or a position is open when first data message arrives"""
        return hasattr(self, "parent_trade") or not self.no_position

    @property
    def position(self) -> int:
        """sets the position associated with the symbol"""
        _position = [i for i in self.app.positions(self.account) if i.contract.symbol == self.contract.symbol]
        return _position[0].position if _position else 0

    @property
    def pnl(self):
        """fetches the PnL associated with the symbol"""

    def log_data_message_to_terminal(self) -> None:
        """logs market data messages to terminal"""

    def persist_history_to_logs(self, data_dir) -> None:
        """sends logs to data directory"""

    def daily_profit_target_reached(self):
        """checks if daily profit threshold is reached"""

    def daily_loss_threshold_reached(self):
        """checks if daily loss threshold is reached"""

    def order_filled_message(self) -> None:
        """logs fills to terminal"""

    def cancel_flanks(self) -> None:
        """used to cancel stops and take-profits"""

    def resubmit_flanks(self) -> None:
        """
        submits new stops and take-profits if cancelled by API
        """
        if self.trailing_order.orderStatus.status in ["Inactive", "Cancelled"]:
            self.place_trailing_order()

        rprint(
            f"[green][RE-FLANKING TRADE][/green]{datetime.datetime.now()}",
            f"ID: {self.parent_trade.order.orderId}",
            f"ACTION: {self.parent_trade.order.action}",
            f"QTY: {self.parent_trade.order.totalQuantity}",
            f"LMT: {self.parent_trade.order.lmtPrice}",
        )

    def place_trailing_order(self, **kwargs) -> None:
        action = "BUY" if self.parent_trade.order.action == "SELL" else "SELL"
        entry_price = self.parent_trade.order.lmtPrice
        limit_price = entry_price - self.offset["stop"] if action == "BUY" else entry_price + self.offset["stop"]
        trailing_order = ib.Order(
            orderType=self.trail_type,
            action=action,
            totalQuantity=self.parent_trade.order.totalQuantity,
            trailStopPrice=limit_price,
            auxPrice=self.aux_price,
            lmtPriceOffset=self.stop_limit_offset,
            **kwargs,
        )

        self.trailing_order = self.app.placeOrder(self.contract, trailing_order)

    def cancel_parentorder_if_stale(self) -> None:
        time_delta = (datetime.datetime.now() - self.timeoftrade).total_seconds()
        if time_delta >= self.time_sentinel:
            if self.parent_trade.isActive():
                rprint(
                    f"[bold green][CANCELLING TRADE {self.parent_trade.order.orderId}][/bold green]",
                    f"{datetime.datetime.now()}",
                )
                if self.parent_trade.order.orderId in [o.orderId for o in self.app.openOrders()]:
                    self.app.cancelOrder(self.trailing_order.order)
                    self.app.cancelOrder(self.parent_trade.order)

    def execute_trade(self) -> None:
        """
        creates an order and places the associated trades

        Example:

        ```python
        action = "BUY" if some_logic_is_true else "SELL"
        entry_price = self.bars[-1].close
        limit_order = ib.LimitOrder(action=action, totalQuantity=1, lmtPrice=limit_price)
        self.trade = self.app.placeOrder(self.contract, limit_order)
        ```

        """

    def trader_logic(self):
        """the trader's execution logic

        Example:

        ```python
        if self.start_trading_ops:
            self.execute_trade(trade_message="opening trade")

        if self.continue_trading_ops:
            if self.pending_parentorder:
                self.cancel_parentorder_if_stale()

            if self.flank_cancelled:
                self.resubmit_flanks()

            if self.reentry:
                self.execute_trade(trade_message="re-entry")

            if self.trend_reversal:
                self.cancel_flanks(purpose="reverse")
                self.execute_trade(trade_message="reversal")
        ```

        """

    async def run_async(self) -> None:
        """
        allows for async streaming

        Example:

        ```python
        with await self.app.connectAsync(
            self.host,
            self.ports[self.platform][self.connection_type],
            clientId=self.clientid,
        ):

            self.bars = self.app.reqRealTimeBars(
                self.contract,
                barSize=5,
                whatToShow="TRADES",
                useRTH=False,
                realTimeBarsOptions=[],
            )

            rprint(f"[bold green][DATA REQUESTED][/bold green] {datetime.datetime.now()}")

            async for update_event in self.bars.updateEvent:
                if self.persist_history:
                    self.update_history()

                log_data_message_to_terminal(
                    self.contract.symbol, self.bars, self.fastma, self.slowma, self.position, self.trend_direction
                )

                self.trader_logic()
        ```

        """

    def run(self):
        """runs non-async jobs

        Example:

        ```python
        self.app.connect(
            self.host,
            self.ports[self.platform][self.connection_type],
            clientId=self.clientid,
        )

        self.bars = self.app.reqRealTimeBars(
            self.contract,
            barSize=5,
            whatToShow="TRADES",
            useRTH=False,
            realTimeBarsOptions=[],
        )

        rprint(f"[bold green][DATA REQUESTED][/bold green] {datetime.datetime.now()}")

        ib.util.sleep(30)

        if self.persist_history:
            self.update_history()

        log_data_message_to_terminal(
            self.contract.symbol, self.bars, self.fastma, self.slowma, self.position, self.trend_direction
        )

        self.trader_logic()
        ```

        """

    def stop(self) -> None:
        self.app.disconnect()
        sys.exit()



# Copyright Justin R. Goheen.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# borrowed from https://github.com/Textualize/rich/blob/master/examples/downloader.py
import os.path
import signal
import sys
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from threading import Event
from typing import Iterable
from urllib.request import urlopen

from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TaskID,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)

from ibquant.types import PATHLIKE

progress = Progress(
    TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
    BarColumn(bar_width=None),
    "[progress.percentage]{task.percentage:>3.1f}%",
    "•",
    DownloadColumn(),
    "•",
    TransferSpeedColumn(),
    "•",
    TimeRemainingColumn(),
)


done_event = Event()


def handle_sigint(signum, frame):
    done_event.set()


signal.signal(signal.SIGINT, handle_sigint)


def copy_url(task_id: TaskID, url: str, path: str) -> None:
    """Copy data from a url to a local file."""
    progress.console.log(f"Requesting {url}")
    response = urlopen(url)
    # This will break if the response doesn't contain content length
    progress.update(task_id, total=int(response.info()["Content-length"]))
    with open(path, "wb") as dest_file:
        progress.start_task(task_id)
        for data in iter(partial(response.read, 32768), b""):
            dest_file.write(data)
            progress.update(task_id, advance=len(data))
            if done_event.is_set():
                return
    progress.console.log(f"Downloaded {path}")


def download(urls: Iterable[str], destination: PATHLIKE):
    """Download multiple files to the given directory."""

    with progress:
        with ThreadPoolExecutor(max_workers=4) as pool:
            for url in urls:
                filename = url.split("/")[-1]
                dest_path = os.path.join(destination, filename)
                task_id = progress.add_task("download", filename=filename, start=False)
                pool.submit(copy_url, task_id, url, dest_path)



# Copyright Justin R. Goheen.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import math

import numpy as np

import ib_insync as ib


def spx_nearest_tick(x):
    """rounds to nearest tick increment"""
    return math.ceil(x * 4) / 4


class EquityFutureFrontMonth:

    _now = datetime.datetime.now()  # gets time-now
    _year = str(_now.year)  # gets the year, makes the int a string
    _month = "0" + str(_now.month) if _now.month < 10 else str(_now.month)  # gets the month, makes the int a string
    _weekday = _now.weekday()  # gets the weekday number (0 = monday, 6 = sunday)
    _dayofmonth = _now.day  # gets the numerical day of month

    _expiry_months = ["03", "06", "09", "12"]  # months in which s&p 500 contracts expire

    for expmonth in _expiry_months:
        if _now.month <= int(
            expmonth
        ):  # sets the expiry month from the list if current month is less than expiry month
            front_month = expmonth
            break  # causes for loop to break once the month is set

    # rolls after week 2 if in expiry month
    _roll = (_month in _expiry_months) and (np.ceil(_dayofmonth / 7) >= 3)

    # rolls to the next expiry is above bool condition is True
    if _roll:
        front_month = _expiry_months[_expiry_months.index(_month) + 1 if int(_month) < 12 else 0]

    expiry = "".join([str(_year), front_month])  # creates a string to pass to ibkr api to get contract


class MarketContracts:

    EquityFutureMinis = dict(
        SNP500=ib.Future("ES", exchange="GLOBEX"),
        RUSSELL=ib.Future("RTY", exchange="GLOBEX"),
        N100=ib.Future("NQ", exchange="GLOBEX"),
    )

    EquityFutureMicros = dict(
        SNP500=ib.Future("MES", exchange="GLOBEX"),
        RUSSELL=ib.Future("M2K", exchange="GLOBEX"),
        N100=ib.Future("MNQ", exchange="GLOBEX"),
    )

    EquityFutureMicrosIndex = dict(
        SNP500=ib.Index("MES", exchange="GLOBEX"),
        RUSSELL=ib.Future("M2K", exchange="GLOBEX"),
        N100=ib.Future("MNQ", exchange="GLOBEX"),
    )

    VixTerms = dict(
        VIX=ib.Index("VIX", exchange="CBOE"),
        VIX3M=ib.Index("VIX3M", exchange="CBOE"),
        VIX6M=ib.Index("VIX6M", exchange="CBOE"),
        VIX1Y=ib.Index("VIX1Y", exchange="CBOE"),
        VIX9D=ib.Index("VIX9D", exchange="CBOE"),
    )

    IndexVolatility = dict(
        SNP500=ib.Index("VIX", exchange="CBOE"),
        N100=ib.Index("VXN", exchange="CBOE"),
        RUSSELL=ib.Index("RVX", exchange="CBOE"),
    )



# Copyright Justin R. Goheen.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import os
import sys

import click
from rich import print as rprint
from rich.prompt import Confirm, Prompt

from ibquant.cli.commands import CLI as cli

IBC_LATEST = "3.14.0"
os.environ["IBC_LATEST"] = IBC_LATEST


@click.group()
def main():
    pass


# ---------------
# main commands
# ---------------
@main.command("login")
def login():
    print()
    os.environ["USER"] = Prompt.ask("Please enter your [green]username[/green]")
    os.environ["PASSWORD"] = Prompt.ask("Please enter your [green]password[/green]")
    print()
    rprint(f"You entered: [green]{os.environ['USER']}[/green] and [green]{os.environ['PASSWORD']}[/green]")
    print()


@main.command("connect")
@click.option(
    "--platform",
    default="tws",
    type=click.Choice(["tws", "gateway"], case_sensitive=True),
    prompt=True,
)
@click.option(
    "--connection-type",
    default="live",
    type=click.Choice(["live", "paper"], case_sensitive=True),
    prompt=True,
)
def ibconnect(platform, connection_type):
    pass


# ---------------
# JVM IB Controller commands
# ---------------
@main.group()
def controller():
    pass


@controller.command("install")
@click.option("--dest", default="ibc")
def get_ibc(dest):
    valid_os = dict(mac="IBCMacos", windows="IBCWin", linux="IBCLinux")
    print()
    opsys = Prompt.ask("Please confirm your operating system", choices=["mac", "windows", "linux"])
    opsys = valid_os[opsys]
    url = f"https://github.com/IbcAlpha/IBC/releases/download/{IBC_LATEST}/{opsys}-{IBC_LATEST}.zip"
    rprint(f"This will install: [green]{url}[/green] to [green]{os.path.join(os.getcwd(), dest)}[/green]")
    rprint("[bold red]The destination directory will be added to the .gitignore[/bold red]")
    confirmed = Confirm.ask("Do you wish to continue?")
    cli.install_controller_if_confirmed(confirmed, url, dest, opsys)
    print()


@controller.command("setup")
def setup_ibc():
    rprint("This will assist with setting up the IBC config")


# ---------------
# advisor commands
# ---------------
@main.group()
def advisor():
    pass


@advisor.command("get-group")
@click.argument("group-name")
@click.option(
    "--platform",
    default="tws",
    type=click.Choice(["tws", "gateway"], case_sensitive=True),
    prompt=True,
)
@click.option(
    "--connection-type",
    default="live",
    type=click.Choice(["live", "paper"], case_sensitive=True),
    prompt=True,
)
def get_group(group_name, platform, connection_type):
    app = cli(platform, connection_type)
    app.connect()
    group = app.get_group_members(group_name)
    app.disconnect()
    cli.rich_table_from_ibiterable(group, field_title="Group Members")


@advisor.command("managed-accounts")
@click.option(
    "--platform",
    default="tws",
    type=click.Choice(["tws", "gateway"], case_sensitive=True),
    prompt=True,
)
@click.option(
    "--connection-type",
    default="live",
    type=click.Choice(["live", "paper"], case_sensitive=True),
    prompt=True,
)
def managed_accounts(platform, connection_type):
    app = cli(platform, connection_type)
    app.connect()
    accounts = app.get_managed_account()
    app.disconnect()
    cli.rich_table_from_ibiterable(accounts, field_title="managed accts")


# ---------------
# account commands
# ---------------
@main.group()
def account():
    pass


@account.command("summary")
@click.option(
    "--platform",
    default="tws",
    type=click.Choice(["tws", "gateway"], case_sensitive=True),
    prompt=True,
)
@click.option(
    "--connection-type",
    default="live",
    type=click.Choice(["live", "paper"], case_sensitive=True),
    prompt=True,
)
@click.option("--account", default="All")
@click.option(
    "--report-type",
    default="summary",
    type=click.Choice(["summary", "values"], case_sensitive=True),
    prompt=True,
)
def account_summary(platform, connection_type, account, report_type):
    account = account.upper() if account.upper() != "ALL" else account.title()
    app = cli(platform, connection_type)
    app.connect()
    summary = app.get_account_report(account, report_type=report_type)
    app.disconnect()
    cli.rich_table_from_ibiterable(summary)


@account.command()
def balance():
    pass


@account.command()
def trades():
    pass


@account.command()
def buying_power():
    pass


@account.command()
def daily_pnl():
    pass


# ---------------
# contract lookup commands
# ---------------
@main.group()
def contract():
    pass


@contract.command("lookup")
@click.option(
    "--platform",
    default="tws",
    type=click.Choice(["tws", "gateway"], case_sensitive=True),
    prompt=True,
)
@click.option(
    "--connection-type",
    default="live",
    type=click.Choice(["live", "paper"], case_sensitive=True),
    prompt=True,
)
@click.option(
    "--contract-type",
    type=click.Choice(["Stock", "Future", "Commodity", "Index", "Fund", "Forex"], case_sensitive=True),
    required=True,
)
@click.option("--con-id", required=False)
def conid_lookup(platform, connection_type, contract_type, con_id):
    app = cli(platform=platform, connection_type=connection_type, contract_type=contract_type)
    app.connect()
    contract_params = [
        i for i in list(inspect.signature(app.contract_method).parameters) if i not in ["args", "kwargs"]
    ]
    kwargs = {k: "" for k in contract_params}
    for param in contract_params:
        if param == "lastTradeDateOrContractMonth":
            msg = "Please enter the expiry month in YYYYMM format"
        else:
            msg = f"Please enter the {param}"
        kwargs[param] = click.prompt(msg, default="", show_default=False)
    contract_details = app.contract_details(**kwargs)
    app.disconnect()
    if not contract_details:
        rprint("[red]See error logs above[/red]")
        sys.exit()
    rprint(f"[green]The contract is {contract_details.contract}[/green]")


# ---------------
# trade commands
# ---------------
@main.group()
def trader():
    pass


@trader.command("trade")
@click.argument("strategy")
@click.argument("market")
@click.argument("contracts")
@click.argument("ordertype")
@click.argument("logger")
def trade(strategy, market, contracts, ordertype, logger):
    pass


@trader.command("plot")
@click.argument("market")
@click.argument("indicators")
def plot(market, indicators):
    pass


# ---------------
# data commands
# ---------------
@main.group()
def data():
    pass


@data.command("fetch-data")
@click.argument("contract")
@click.argument("end-date")
@click.argument("duration")
@click.argument("barsize")
@click.argument("show")
@click.argument("use-rth")
@click.argument("dateformat")
@click.argument("keep-updated")
def fetch_data():
    pass


# ---------------
# data commands
# ---------------
@main.group()
def researcher():
    pass


@researcher.command("run")
@click.argument("max-drawdown")
@click.argument("long-short")
@click.argument("execution-window")
def run_backtest(maxdd, long_short, execution_window):
    pass


# ---------------
# performance commands
# ---------------

# ---------------
# zipline commands
# https://zipline.ml4trading.io/beginner-tutorial.html#command-line-interface
# ---------------

# ---------------
# pyfolio commands
# ---------------

# ---------------
# alphalens commands
# ---------------

# ---------------
# empyrical commands
# ---------------

# ---------------
# quandl commands
# ---------------



# Copyright Justin R. Goheen.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import plotly.graph_objects as go


def plot_strategy(data, symbol="SPY"):
    """plots strategy returns given a dataframe"""

    trading = data.loc[data["Trades"] != 0, "Trades"]
    starts = trading.index[::2]
    stops = trading.index.difference(starts)

    fig = go.Figure()

    fig.add_trace(go.Scatter(x=data.index, y=data["Close"], name=f"{symbol} Close"))

    for start, stop in zip(starts, stops):
        fig.add_vrect(
            x0=start,
            x1=stop,
            fillcolor="grey",
            opacity=0.25,
            line_width=0,
        )

    fig.update_layout(
        title={
            "text": symbol + " (grey shaded areas indicate flat position)",
            "y": 0.97,
            "x": 0.5,
            "xanchor": "center",
            "yanchor": "top",
        },
        legend=dict(title="", yanchor="top", y=1.06, xanchor="left", x=0, orientation="h"),
        xaxis_rangeslider_visible=False,
        height=400,
        width=1000,
        margin=dict(l=25, r=10, b=25, t=30),
    )

    return fig



# Copyright Justin R. Goheen.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import time
from abc import ABC
from typing import Optional

import numpy as np
import pandas as pd
from rich import print as rprint

import ib_insync as ib
from ibquant.futures import EquityFutureFrontMonth


class DataMixins(ABC):
    def __init__(self) -> None:
        super().__init__()

    def historical_bars(
        self,
        end_date_time: str = "",
        duration: str = "1 D",
        bar_size: str = "1 min",
        what_to_show: str = "TRADES",
        use_rth: bool = False,
    ):
        """fetches historical bars"""

        self.app.connect(
            self.host,
            self.ports[self.platform][self.connection_type],
            clientId=self.clientid,
        )

        data = self.app.reqHistoricalData(
            self.contract,
            endDateTime=end_date_time,
            durationStr=duration,
            barSizeSetting=bar_size,
            whatToShow=what_to_show,
            useRTH=use_rth,
        )

        ib.util.sleep(30)
        self.app.disconnect()
        ib.util.sleep(5)

        return data

    def _process_bars(self):
        """used to process bars to make dataframe"""
        pass

    def historical_ticks(
        self,
        start_date_time: str = "20210319 09:30:00",
        end_date_time: str = "",
        number_of_ticks: int = 1000,  # max is 1000
        what_to_show: str = "TRADES",  # TRADES, MIDPOINT, BID, ASK, BID_ASK
        ignore_size: bool = False,
        use_rth: bool = True,
    ):
        """fetches historical ticks"""

        data = self.app.reqHistoricalTicks(
            self.contract,
            startDateTime=start_date_time,
            endDateTime=end_date_time,
            numberOfTicks=number_of_ticks,
            whatToShow=what_to_show,
            useRth=use_rth,
            ignoreSize=ignore_size,
        )

        time.sleep(60)

        return data

    def _process_ticks(self):
        """used to process ticks for dataframe"""
        pass

    def histogram(
        self,
        use_rth: bool = False,
        period: str = "15 min",
    ):
        """
        fetches a historical histogram of volume at price

        Note:
            https://ib-insync.readthedocs.io/api.html#ib_insync.ib.IB.reqHistoricalTicks
            http://interactivebrokers.github.io/tws-api/historical_time_and_sales.html#reqHistoricalTicks
        """

        data = self.app.reqHistogramData(
            self.contract,
            use_rth,
            period,
        )

        rprint(f"[bold green][FETCHING HISTOGRAM][/bold green] {datetime.datetime.now()}")

        ib.util.sleep(60)

        return data

    def news(self):
        pass

    def to_dataframe(self, data):
        return ib.util.df(data)



# Copyright Justin R. Goheen.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os


def ignore_path(dest, dest_is_dir=True):
    if dest_is_dir:
        dest = dest + os.sep
    if ".gitignore" not in os.listdir(os.getcwd()):
        with open(os.path.join(os.getcwd(), ".gitignore"), "a") as file:
            file.write("\n# added by ibtrader")
            file.write(f"\n{dest}")
        file.close()
    else:
        with open(os.path.join(os.getcwd(), ".gitignore"), "a") as file:
            file.write("\n# added by ibtrader")
            file.write(f"\n{dest}")
        file.close()



# Copyright Justin R. Goheen.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import os
from importlib import import_module
from typing import Any, List, Optional, Union

from rich import print as rprint
from rich.console import Console
from rich.table import Table

import ib_insync as ib
from ib_insync.objects import AccountValue
from ibquant.core import AppBase
from ibquant.io import add_ibconfigs_section, download, ignore_path, unzip
from ibquant.types import PATHLIKE


class CLI(AppBase):
    def __init__(self, *args: Any, **kwargs: Any):
        super(CLI, self).__init__(*args, **kwargs)

    @staticmethod
    def install_controller_if_confirmed(confirmed: bool, url: str, destination: PATHLIKE, opsys: str) -> None:

        if confirmed:
            ignore_path(destination, dest_is_dir=True)

            if not os.path.isdir(destination):
                os.mkdir(destination)

            download_destination = os.path.join(os.getcwd(), destination)

            download(
                urls=[url],
                destination=download_destination,
            )

            filepath = os.path.join(download_destination, f"{opsys}-{os.environ['IBC_LATEST']}.zip")

            unzip(filepath, download_destination)

            add_ibconfigs_section(os.path.join(download_destination, "config.ini"))

        if not confirmed:
            rprint("[bold red]Exiting[/bold red]")

    @staticmethod
    def rich_table_from_ibiterable(
        ibiterable: Union[List[str], List[AccountValue]], title: Optional[str] = None, **kwargs: Any
    ) -> None:

        table = Table(title=title)

        if not isinstance(ibiterable, list):
            raise TypeError(f"must supply a list, got {type(ibiterable)}")

        if hasattr(ibiterable[0], "_fields"):
            fields = ibiterable[0]._fields
            for field in fields:
                table.add_column(field, justify="right", no_wrap=True)
            for record in ibiterable:
                record = record._asdict()
                table.add_row(*[record.get(field) for field in fields])

        if isinstance(ibiterable[0], str):
            table.add_column(kwargs["field_title"], justify="right", no_wrap=True)
            for record in ibiterable:
                table.add_row(record)

        console = Console()
        console.print(table)

    @staticmethod
    def inspect_signature(modulename, attrname):
        module = import_module(modulename)
        moduleclass = getattr(module, attrname)
        return inspect.signature(moduleclass)



# Copyright Justin R. Goheen.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC


class AccountMixins(ABC):
    def __init__(self):
        super().__init__()

    def get_account_report(self, account, report_type="summary"):
        reports = dict(values=self.app.accountValues, summary=self.app.accountSummary)
        report = reports[report_type]
        account_report = report()
        if report_type == "summary":
            account_report = [i for i in account_report if i.account == account]
        return account_report

    def get_managed_account(self):
        accounts = self.app.managedAccounts()
        return accounts



# Copyright Justin R. Goheen.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from typing import Any, Callable, Dict, List, Union

import ib_insync as ib
from ibquant.exceptions import ContractException


class ContractMixins(ABC):
    def __init__(self) -> None:
        super().__init__()

    @property
    def contractdata(self) -> Dict[Callable, str]:
        return getattr(Contracts, self.contract_type)

    @property
    def contract_method(self) -> Callable:
        return self.contractdata["method"]

    @property
    def contract_type(self):
        """used for sandbox traders"""
        return self._contract_type

    @contract_type.setter
    def contract_type(self, contract_type: str):
        self._contract_type = contract_type

    @property
    def sectype(self) -> str:
        return self.contractdata["sectype"]

    def qualify_contract(self, **kwargs) -> Union[List, List[ib.Contract]]:
        return self.app.qualifyContracts(self.contract_method(**kwargs))

    def contract_details(self, *args: Any, **kwargs: Any) -> ib.Contract:
        qc = self.qualify_contract(**kwargs)
        if not qc:
            return None
        _details = self.app.reqContractDetails(self.contract_method(**kwargs))
        return _details[0]

    def validate_contract_config(self):
        """validates that required params exist"""


class Contracts:
    Contract = {"method": ib.contract.Contract, "sectype": ""}
    Stock = {"method": ib.contract.Stock, "sectype": "STK"}
    Option = {"method": ib.contract.Option, "sectype": "OPT"}
    Future = {"method": ib.contract.Future, "sectype": "FUT"}
    ContFuture = {"method": ib.contract.ContFuture, "sectype": "CONTFUT"}
    Forex = {"method": ib.contract.Forex, "sectype": "CASH"}
    Index = {"method": ib.contract.Index, "sectype": "IND"}
    CFD = {"method": ib.contract.CFD, "sectype": "CFD"}
    Bond = {"method": ib.contract.Bond, "sectype": "BOND"}
    Commodity = {"method": ib.contract.Commodity, "sectype": "CMDTY"}
    FuturesOption = {"method": ib.contract.FuturesOption, "sectype": "FOP"}
    MutualFund = {"method": ib.contract.MutualFund, "sectype": "FUND"}
    Warrant = {"method": ib.contract.Warrant, "sectype": "WAR"}
    IOPT = {"method": ib.contract.Warrant, "sectype": "IOPT"}
    Bag = {"method": ib.contract.Bag, "sectype": "BAG"}
    Crypto = {"method": ib.contract.Crypto, "sectype": "CRYPTO"}
    News = {"method": ib.contract.Contract, "sectype": "NEWS"}



# Copyright Justin R. Goheen.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import plotly.graph_objects as go
from plotly.colors import diverging
from plotly.subplots import make_subplots


class MarketDepthVisualizations:
    def plot_spread(data, index_location=1):
        """plots bids on right and asks on left to visualize mid price as spread"""

        prices = [i for i in data.columns if "price" in i]
        sizes = [i for i in data.columns if "size" in i]

        fig = make_subplots(specs=[[{"secondary_y": True}]])
        fig.add_trace(
            go.Scatter(
                x=data.iloc[index_location][[i for i in prices if "bid" in i]],
                y=data.iloc[index_location][[i for i in sizes if "bid" in i]],
                fill="tozeroy",
                name="Bids",
            )
        )
        fig.add_trace(
            go.Scatter(
                x=data.iloc[index_location][[i for i in prices if "ask" in i]],
                y=data.iloc[index_location][[i for i in sizes if "ask" in i]],
                fill="tozeroy",
                name="Asks",
            ),
            secondary_y=True,
        )
        fig.update_layout(
            title={
                "text": "Spread",
                "y": 0.97,
                "x": 0.5,
                "xanchor": "center",
                "yanchor": "top",
            },
            legend=dict(title="", yanchor="top", y=1.06, xanchor="left", x=0, orientation="h"),
            xaxis_rangeslider_visible=False,
            height=400,
            width=1000,
            margin=dict(l=25, r=10, b=25, t=30),
        )
        fig.show()

    def plot_levels(data, levels=10):
        """
        plots a cmap-like timeseries plot to visualize dominate bid and asks across time
        levels: max number is 10
        """
        fig = go.Figure()
        for i in range(1, levels + 1):
            fig.add_trace(go.Scatter(y=data[f"bidprice{i}"], marker_color=diverging.Spectral[i], showlegend=False))
            fig.add_trace(go.Scatter(y=data[f"askprice{i}"], marker_color=diverging.Spectral[i], showlegend=False))

        fig.update_layout(
            title={
                "text": "Levels",
                "y": 0.97,
                "x": 0.5,
                "xanchor": "center",
                "yanchor": "top",
            },
            legend=dict(title="", yanchor="top", y=1.06, xanchor="left", x=0, orientation="h"),
            xaxis_rangeslider_visible=False,
            height=400,
            width=1000,
            margin=dict(l=25, r=10, b=25, t=30),
        )

        fig.show()

    def plot_heatmap(data, levels=5):
        """overlays a mid-price plot onto a heatmap of dominate bids and asks"""

        prices = [i for i in data.columns if "price" in i]
        all_sizes = [i for i in data.columns if "size" in i]
        bids = [i for i in all_sizes if "bid" in i]
        asks = [i for i in all_sizes if "ask" in i]
        sizes = asks[::-1][: levels + 1] + bids[: levels + 1]  # make sizes meet in the middle

        fig = make_subplots(specs=[[{"secondary_y": True}]])

        fig.add_trace(
            go.Scatter(y=data[prices].mean(axis=1), marker_color="white", line=dict(width=4)), secondary_y=True
        )
        fig.add_trace(go.Heatmap(z=data[sizes].T, name="Bid Ask Size"))

        fig.update_layout(
            title={
                "text": str(data.index[0].date())
                + f" start: {str(data.index[0].time())}"
                + f" end: {str(data.index[-1].time())}",
                "y": 0.97,
                "x": 0.2,
                "xanchor": "left",
                "yanchor": "top",
            },
            xaxis_rangeslider_visible=False,
            height=400,
            width=1000,
            margin=dict(l=25, r=10, b=25, t=30),
        )

        fig.show()



# Copyright Justin R. Goheen.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

from setuptools import find_packages, setup

rootdir = Path(__file__).parent
long_description = (rootdir / "README.md").read_text()

setup(
    name="ibquant",
    version="0.1",
    description="A Python Framework for Interactive Brokers TWS API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Justin Goheen",
    license="Apache 2.0",
    author_email="",
    zip_safe=False,
    classifiers=[
        "Environment :: Console",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Office/Business :: Financial :: Investment",
        "Topic :: Software Development :: User Interfaces",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.10",
    ],
)


# python = ">=3.10, <3.12"



from ibquant.mixins.account import AccountMixins
from ibquant.mixins.advisor import AdvisorMixins
from ibquant.mixins.connect import ConnectionMixins
from ibquant.mixins.contract import ContractMixins
from ibquant.mixins.data import DataMixins



# Copyright Justin R. Goheen.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from abc import ABC
from typing import Tuple

import ib_insync as ib
from ibquant.types import PATHLIKE

ACCOUNT_TYPE = "IB.accountSummary"


class SimpleLogger(ABC):
    def __init__(self, starting_capital: Tuple[float], account: ACCOUNT_TYPE):
        super().__init__()
        self.starting_capital = starting_capital
        self.account = account

    def returns(self):
        return self.account.value - self.starting_capital

    def persist_history_to_logs(self, data_dir: PATHLIKE, filename: str, filetype: str) -> None:
        if filetype == "pq":
            filetype = "parquet"
        persist_method = getattr(self.history, f"to_{filetype}")
        persist_method("".join([data_dir, filename, ".", filetype]))



# Copyright Justin R. Goheen.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import ib_insync as ib
from ibquant.mixins import AccountMixins, AdvisorMixins, ConnectionMixins, ContractMixins, DataMixins


class AppBase(AccountMixins, AdvisorMixins, ConnectionMixins, ContractMixins, DataMixins):
    def __init__(
        self, platform: str, connection_type: str, account: Optional[str] = None, contract_type: Optional[str] = None
    ):
        super().__init__()
        self._app = ib.IB()
        self.platform = platform
        self.connection_type = connection_type
        self.contract_type = contract_type
        self.account = account

    @property
    def app(self):
        return self._app
